// ----------------------------------------------------------------------------
// Project:	Implementation Alternatives for Bottom-Up Evaluation
// Version:	Submission to ICLP 2010
// Filename:	test7push.cpp
// Purpose:	Time Measurements for Rule Implementations: Push Method
// Last Change:	30.01.2010
// Language:	C++
// EMail:	brass@acm.org
// WWW:		http://www.informatik.uni-halle.de/~brass/
// Address:	Univ. Halle, Von-Seckendorff-Platz 1, D-06120 Halle, GERMANY
// Copyright:	(c) 2010 by Stefan Brass
//		Downloads and modifications for tests and scientific purposes
//		permitted, provided that some acknowledgement of my work
//		remains, and modified versions are clearly marked as such.
//		There is no warranty at all - this code may contain bugs.
// ----------------------------------------------------------------------------

// Test Program 7:
//
// p(X, Z, 2) :- q(X, Y), r(Y, 5, Z).
// q(X, Y) :- s(X, Y), Y >= 0.
//
// r = {(i,j,k) | 1 <= i <= 10000, 1 <= j <= 100, 1 <= k <= 100}
//	r can be accessed with binding pattern bff (index on first argument)
// s = {(i,i) | 1 <= i <= 10000 }


#include <iostream>
using namespace std;

#if defined(_MSC_VER) || defined(_WINDOWS_)
#define WINDOWS
#else
#define UNIX
#endif

// ----------------------------------------------------------------------------
// Sizes for test:
// ----------------------------------------------------------------------------

const int S_TUPLES = 10000;

const int R_COL1_VALUES = 10000;

const int R_COL2_VALUES = 100;

const int R_COL3_VALUES = 100;

const bool PRINT_EXTENSION = false;

const int STACKSIZE = 50;

// ----------------------------------------------------------------------------
// Timing Functions (UNIX Version):
// ----------------------------------------------------------------------------

#ifdef UNIX
#include <sys/time.h>
#include <time.h>

class timer {
public:
	void start()
	{
		gettimeofday(&Time_Start, 0);
		Clock_Start = clock();
	}

	void end()
	{
		gettimeofday(&Time_End, 0);
		Clock_End = clock();
	}
	
	long int realtime_msec()
	{
		long sec = Time_End.tv_sec - Time_Start.tv_sec;
		long usec = Time_End.tv_usec - Time_Start.tv_usec;
		if(usec < 0) {
			usec += 1000000;
			sec -= 1;
		}
		return sec * 1000 + (usec + 500) / 1000;
	}
	long int cputime_msec()
	{
		long ticks = Clock_End - Clock_Start;
		return ticks / (CLOCKS_PER_SEC / 1000);
	}

private:
	timeval Time_Start, Time_End;
	clock_t Clock_Start, Clock_End;
};
#endif

// ----------------------------------------------------------------------------
// Timing Functions (Windows Version):
// ----------------------------------------------------------------------------

#ifdef WINDOWS
#include <windows.h>
#include <time.h>

class timer {
public:
	void start()
	{
		GetSystemTimeAsFileTime(&(Start.ft));
		Clock_Start = clock();
	}

	void end()
	{
		GetSystemTimeAsFileTime(&(End.ft));
		Clock_End = clock();
	}
	
	long int realtime_msec()
	{
		return static_cast<long> ((End.ns100 - Start.ns100) / 10000);
	}
	long int cputime_msec()
	{
		long ticks = Clock_End - Clock_Start;
		return ticks / (CLOCKS_PER_SEC / 1000);
	}

private:
	union {
		long long ns100;
		FILETIME ft;
	} Start, End;
	clock_t Clock_Start, Clock_End;
};
#endif


// ----------------------------------------------------------------------------
// Stack (of int):
// ----------------------------------------------------------------------------


class stack {

private:
	static int stack_arr[STACKSIZE];
	static int *stack_p;
	static int stack_free;

public:
	static inline void stack_init()
	{
		stack_p = stack_arr;
		stack_free = STACKSIZE;
	}

	static inline void push(int i)
	{
		if(stack_free > 0) {
			stack_free--;
			*(stack_p++) = i;
		}
		else {
			cout << "Stack overflow\n";
			exit(1);
		}
	}

	static inline int pop()
	{
		if(stack_p > stack_arr) {
			stack_free++;
			return *(--stack_p);
		}
		else {
			cout << "Stack underflow\n";
			exit(1);
		}
	}

	static inline bool is_empty()
	{
		return (stack_p == stack_arr);
	}
};

int stack::stack_arr[STACKSIZE];
int *stack::stack_p;
int stack::stack_free;

//-----------------------------------------------------------------------------
// Predicate s: Tuples generated by computation
//-----------------------------------------------------------------------------

class s_cursor {
public:
	s_cursor() { }

	void open() { i = 0; }

	bool fetch()
	{
		if(i < S_TUPLES) {
			i++;
			return true;
		}
		else
			return false;
	}

	int col_1() { return i; }

	int col_2() { return i; }

	void close() {}

	void push() { stack::push(i); }

	void pop() { i = stack::pop(); }
private:
	int i;
};

//-----------------------------------------------------------------------------
// Predicate r (binding pattern bff): Tuples generated by computation
//-----------------------------------------------------------------------------

class r_cursor_bff {
public:
	r_cursor_bff() { }

	void open(int c1)
	{
		i = c1;
		if(i > 0 && i <= R_COL1_VALUES) {
			j = 0;
			k = R_COL3_VALUES;
		}
		else {
			j = R_COL2_VALUES + 1;
			k = 0;
		}
	}

	bool fetch()
	{
		if(j <= R_COL2_VALUES && k < R_COL3_VALUES) {
			k++;
			return true;
		}
		else if(j < R_COL2_VALUES) {
			j++;
			k = 1;
			return true;
		}
		else
			return false;
	}

	int col_1() { return i; }

	int col_2() { return j; }

	int col_3() { return k; }

	void close() {}

	void push()
	{
		stack::push(i);
		stack::push(j);
		stack::push(k);
	}

	void pop()
	{
		k = stack::pop();
		j = stack::pop();
		i = stack::pop();
	}

private:
	int i;
	int j;
	int k;
};


// ----------------------------------------------------------------------------
// Predicate p - Push Method
// ----------------------------------------------------------------------------


class p_cursor {
private:

	// Continuation labels:
	typedef enum {CONT_q, CONT_init, CONT_init1 } cont_t;

	// Functions for using the global stack:
	inline void push_int(int i)
	{
		stack::push(i);
	}

	inline int pop_int()
	{
		return stack::pop();
	}

	inline void push_cont(cont_t t)
	{
		stack::push(t);
	}

	inline cont_t pop_cont()
	{
		return (cont_t) stack::pop();
	}

public:
	p_cursor() { };

	void open()
	{
		stack::stack_init();
		push_cont(CONT_init);
	}

    bool fetch()
    {
    s_cursor s1;
    r_cursor_bff r1;
    int q_1, q_2;

backtrack:
    if(stack::is_empty())
	    return false;
    switch(pop_cont()) {
    case CONT_init: goto cont_init;
    case CONT_init1: goto cont_init1;
    case CONT_q: goto cont_q;
    default: cout << "Impossible cont\n"; return false;
    }

cont_init:
    s1.open();
    while(s1.fetch()) {
        if(s1.col_2() >= 0) {
            q_1 = s1.col_1();
            q_2 = s1.col_2();
            if(s1.fetch()) {
                s1.push();
                push_cont(CONT_init1);
            }
            goto q;
        }
    }
    return false; // if this is the last/only initialization rule
cont_init1:
    s1.pop();
    do {
        if(s1.col_2() >= 0) {
            q_1 = s1.col_1();
            q_2 = s1.col_2();
            if(s1.fetch()) {
                s1.push();
                push_cont(CONT_init1);
            }
            goto q;
        }
    } while(s1.fetch());
    return false; // if this is the last/only initialization rule

q:  r1.open(q_2); // Assuming index on first arg
    while(r1.fetch()) {
        if(r1.col_2() == 5) {
            p_1 = q_1;
            p_2 = r1.col_3();
            p_3 = 2;
            if(r1.fetch()) {
                push_int(q_1);
                push_int(q_2);
		r1.push();
                push_cont(CONT_q);
            }
            goto p;
        }
    }
    goto backtrack; // if this is the last place where q is used
cont_q:
    r1.pop();
    q_2 = pop_int();
    q_1 = pop_int();
    do {
        if(r1.col_2() == 5) {
            p_1 = q_1;
            p_2 = r1.col_3();
            p_3 = 2;
            if(r1.fetch()) {
                push_int(q_1);
                push_int(q_2);
		r1.push();
                push_cont(CONT_q);
            }
            goto p;
        }
    } while(r1.fetch());
    goto backtrack; // if this is the last place where q is used

p:  return true;
    }

    int col_1() { return p_1; }
    int col_2() { return p_2; }
    int col_3() { return p_3; }

    void close()
    {
    }
private:
	// Arguments:
	int p_1, p_2, p_3;
};


//-----------------------------------------------------------------------------
// Main Program:
//-----------------------------------------------------------------------------

int main()
{
	timer t;
	int c1 = 0;
	int c2 = 0;
	int c3 = 0;

	cout << "Test of Method 3: Push-Method (Program 7)\n";
	t.start();
	p_cursor p1;
	p1.open();
	while(p1.fetch()) {
		c1 += p1.col_1();
		c2 += p1.col_2();
		c3 += p1.col_3();
	}
	p1.close();
	t.end();

	if(PRINT_EXTENSION) {
		cout << "Extension of p:\n\n";
		p1.open();
		int tup = 0;
		while(p1.fetch() && tup++ < 10) {
			cout << "p(" << p1.col_1() << ", " << p1.col_2()
				<< ", " << p1.col_3() << ").\n";
		}
		p1.close();
	}

	cout << "Column sums: col1=" << c1 << ", col2=" << c2 << ", col3="
		<< c3 << "\n\n";
	cout << "\nUsed time: " << t.realtime_msec() << " ms,  "
		<< "CPU: " << t.cputime_msec() << " ms.\n";

	return 0;
}

